/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/



#include "mpris1.h"
#include "mpris_common.h"

#include "EnginePlayer/engineplayer.h"
#include "Widget/audiocontrols.h"
#include "MediaItem/mediaitem.h"

#include "global_actions.h"

#include "utilities.h"
#include "constants.h"

#include "debug.h"

#include <QCoreApplication>
#include <QDBusConnection>
#include <QDBusMetaType>
#include <QtDBus>


/*******************************************************************************
    MPRIS SPECIFICATION
--------------------------------------------------------------------------------
      http://www.mpris.org/1.0/spec.html#org.freedesktop.MediaPlayer
*******************************************************************************/



/*
********************************************************************************
*                                                                              *
*    Mpris1                                                                    *
*                                                                              *
********************************************************************************
*/
Mpris1::Mpris1(QObject* parent) : QObject(parent)
{
    //Debug::warning() << "##### MPRIS1 create";

    m_root      = new Mpris1Root(this);
    m_player    = new Mpris1Player(this);
    m_tracklist = new Mpris1TrackList(this);

    QDBusConnection::sessionBus().registerObject(QLatin1String("/"), m_root, QDBusConnection::ExportAllContents);
    QDBusConnection::sessionBus().registerObject(QLatin1String("/Player"), m_player, QDBusConnection::ExportAllContents);
    QDBusConnection::sessionBus().registerObject(QLatin1String("/TrackList"), m_tracklist, QDBusConnection::ExportAllContents);
    QDBusConnection::sessionBus().registerService("org.mpris.yarock");
}


Mpris1::~Mpris1()
{
    QDBusConnection::sessionBus().unregisterService("org.mpris.yarock");
}


QVariantMap Mpris1::GetMetadata(const MEDIA::TrackPtr track)
{
    QVariantMap ret;
    if(track->type() == TYPE_TRACK) {
      AddMetadata("location",    track->url,                &ret);
      AddMetadata("artist",      track->artist,             &ret);
      AddMetadata("title",       track->title,              &ret);
      AddMetadata("album",       track->album,              &ret);
      AddMetadata("time",        track->duration,           &ret);  // The duration in seconds
      //AddMetadata("mtime",       track->data.duration*1000, &ret); //The duration in milliseconds
      AddMetadata("tracknumber", QString::number(track->num), &ret);
      AddMetadata("year",        QString::number(track->year), &ret);
      AddMetadata("genre",       track->genre, &ret);

      //AddMetadata("arturl",      art_uri, &last_metadata_);
      const QString coverpath = UTIL::CONFIGDIR + "/albums/" + track->coverName();
      if( QFile(coverpath).exists() ) {
        AddMetadata("arturl",      coverpath,  &ret);
      }
      else {
        //TODO
      }
    }
    else  if(track->type() == TYPE_STREAM) {
     //! TODO
    }

    return ret;
}

/*
********************************************************************************
*                                                                              *
*    Mpris1Root                                                                *
*                                                                              *
********************************************************************************
*/
Mpris1Root::Mpris1Root(QObject* parent) : QObject(parent)
{
    qDBusRegisterMetaType<Version>();
}


QString Mpris1Root::Identity()
{
    return QString("%1 %2").arg(APP_NAME,VERSION);
}


Version Mpris1Root::MprisVersion()
{
    Version version;
    version.major = 1;
    version.minor = 0;
    return version;
}

void Mpris1Root::Quit()
{
    qApp->quit();
}

/*
********************************************************************************
*                                                                              *
*    Mpris1Player                                                              *
*                                                                              *
********************************************************************************
*/
Mpris1Player::Mpris1Player(QObject* parent) : QObject(parent)
{
    qDBusRegisterMetaType<DBusStatus>();
    QObject::connect(RepeatControl::instance(), SIGNAL(repeatStateChange(int)), SLOT(RepeatModeChanged()));
    QObject::connect(ShuffleControl::instance(), SIGNAL(shuffleStateChange(int)), SLOT(ShuffleModeChanged()));
    QObject::connect(EnginePlayer::instance(), SIGNAL(engineStateChanged()), SLOT(EngineStateChanged()));
    QObject::connect(EnginePlayer::instance(), SIGNAL(playingInfoChanged()), SLOT(CurrentSongChanged()));
}


void Mpris1Player::Pause()
{
    ACTIONS()->value(ENGINE_PLAY)->trigger();
}

void Mpris1Player::Stop()
{
    ACTIONS()->value(ENGINE_STOP)->trigger();
}

void Mpris1Player::Prev()
{
    ACTIONS()->value(ENGINE_PLAY_PREV)->trigger();
}

void Mpris1Player::Play()
{
    ACTIONS()->value(ENGINE_PLAY)->trigger();
}

void Mpris1Player::Next()
{
    ACTIONS()->value(ENGINE_PLAY_NEXT)->trigger();
}

void Mpris1Player::CurrentSongChanged()
{
    m_last_metadata = static_cast<Mpris1*>(parent())->GetMetadata(EnginePlayer::instance()->playingTrack());

    emit TrackChange(m_last_metadata);
    emit StatusChange(GetStatus());
    emit CapsChange(GetCaps());
}

void Mpris1Player::EngineStateChanged()
{
    emit StatusChange(GetStatus());
    emit CapsChange(GetCaps());
}

void Mpris1Player::ShuffleModeChanged()
{
    emit StatusChange(GetStatus());
}

void Mpris1Player::RepeatModeChanged()
{
    emit StatusChange(GetStatus());
}

void Mpris1Player::Repeat(bool repeat)
{
    //! bool enable : {Repeat_Off = 0, Repeat_Track = 1}
    //! class RepeatControl ==> enum RepeatMode {RepeatOff = 0,RepeatTrack = 1, RepeatAll = 2 };
    RepeatControl::instance()->setState(repeat);
}

DBusStatus Mpris1Player::GetStatus() const
{
    DBusStatus status;

    switch (EnginePlayer::instance()->engineState()) {
      case ENGINE::PLAYING : status.play = DBusStatus::Mpris_Playing; break;
      case ENGINE::PAUSED  : status.play = DBusStatus::Mpris_Paused;  break;
      case ENGINE::STOPPED : status.play = DBusStatus::Mpris_Stopped; break;
      case ENGINE::ERROR   : status.play = DBusStatus::Mpris_Stopped; break;
    }

    status.random          = (ShuffleControl::instance()->getState()) == 1 ? true : false;
    status.repeat_track    = (RepeatControl::instance()->getState()) == 1 ? true : false;
    status.repeat_playlist = (RepeatControl::instance()->getState()) == 2 ? true : false;

    return status;
}


void Mpris1Player::VolumeSet(int volume)
{
    EnginePlayer::instance()->setVolume(volume); // percent volume for player interface
}

int Mpris1Player::VolumeGet() const
{
    return EnginePlayer::instance()->volume(); // player return percent value
}

void Mpris1Player::VolumeUp(int change)
{
    VolumeSet(VolumeGet() + change);
}

void Mpris1Player::VolumeDown(int change)
{
    VolumeSet(VolumeGet() - change);
}

void Mpris1Player::Mute()
{
    EnginePlayer::instance()->setMuted( EnginePlayer::instance()->isMuted() );
}

void Mpris1Player::PositionSet(int pos_msec)
{
    EnginePlayer::instance()->seek(pos_msec);
}

int Mpris1Player::PositionGet() const
{
    return EnginePlayer::instance()->mediaObject()->currentTime();
}

QVariantMap Mpris1Player::GetMetadata() const
{
    return m_last_metadata;
}

int Mpris1Player::GetCaps() const
{
    int capabilities =
      (1 << 0) | // CAN_GO_NEXT
      (1 << 1) | // CAN_GO_PREV
      (1 << 3) | // CAN_PLAY
      (1 << 4) | // CAN_SEEK
      (1 << 5) | // CAN_PROVIDE_METADATA
      (1 << 6);  // CAN_HAS_TRACKLIST

    if( EnginePlayer::instance()->engineState() == ENGINE::PLAYING ) {
      capabilities |= (1 << 2); // CAN_PAUSE
    }

    return capabilities;
}


/*
********************************************************************************
*                                                                              *
*    Mpris1TrackList                                                           *
*                                                                              *
********************************************************************************
*/
Mpris1TrackList::Mpris1TrackList(QObject* parent) : QObject(parent)
{
    //connect(player->playlists(), SIGNAL(PlaylistChanged(Playlist*)), SLOT(PlaylistChanged(Playlist*)));
}

void Mpris1TrackList::PlaylistChanged()
{
    //emit TrackListChange(playlist->rowCount());
}

int Mpris1TrackList::AddTrack(const QString& track, bool play)
{
  Q_UNUSED(track)
  Q_UNUSED(play)
  /*
  player_->playlists()->active()->InsertUrls(
        QList<QUrl>() << QUrl(track), play);
        */
  return 0;
}

void Mpris1TrackList::DelTrack(int index)
{
  Q_UNUSED(index)
  /*
  player_->playlists()->active()->removeRows(index, 1);
  */
}

int Mpris1TrackList::GetCurrentTrack() const
{
  /*
  return player_->playlists()->active()->current_row();

  */
  return 0;
}

int Mpris1TrackList::GetLength() const
{
  /*
  return player_->playlists()->active()->rowCount();
  */
  return 0;
}

QVariantMap Mpris1TrackList::GetMetadata(int pos) const
{
  Q_UNUSED(pos)

  /*
  PlaylistItemPtr item = player_->GetItemAt(pos);
  if (!item)
    return QVariantMap();

  return Mpris1::GetMetadata(item->Metadata());
  */
  return QVariantMap();
}

void Mpris1TrackList::SetLoop(bool enable)
{
    //! bool enable : {Repeat_Off = 0, Repeat_Playlist = 1}
    //! class RepeatControl ==> enum RepeatMode {RepeatOff = 0,RepeatTrack = 1, RepeatAll = 2 };
    RepeatControl::instance()->setState(enable ? 2 : 0);
}

void Mpris1TrackList::SetRandom(bool enable)
{
    //! bool enable : {Shuffle_Off = 0, Shuffle_All = 1}
    //! class ShuffleControl ==>     enum ShuffleMode {ShuffleOff = 0 ,ShuffleOn = 1};
    ShuffleControl::instance()->setState(enable);
}

void Mpris1TrackList::PlayTrack(int index)
{
    Q_UNUSED(index)

/*  player_->PlayAt(index, Engine::Manual, true);
  */
}



/*******************************************************************************
    Dbus operator
*******************************************************************************/
QDBusArgument& operator<< (QDBusArgument& arg, const Version& version) {
    arg.beginStructure();
    arg << version.major << version.minor;
    arg.endStructure();
    return arg;
}

const QDBusArgument& operator>> (const QDBusArgument& arg, Version& version) {
    arg.beginStructure();
    arg >> version.major >> version.minor;
    arg.endStructure();
    return arg;
}

QDBusArgument& operator<< (QDBusArgument& arg, const DBusStatus& status) {
    arg.beginStructure();
    arg << status.play;
    arg << status.random;
    arg << status.repeat_track;
    arg << status.repeat_playlist;
    arg.endStructure();
    return arg;
}

const QDBusArgument& operator>> (const QDBusArgument& arg, DBusStatus& status) {
    arg.beginStructure();
    arg >> status.play;
    arg >> status.random;
    arg >> status.repeat_track;
    arg >> status.repeat_playlist;
    arg.endStructure();
    return arg;
}
