/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/

#include "threadmanager.h"
#include "CollectionModel/collectionpopulator.h"
#include "CollectionModel/plsfilepopulator.h"
#include "Database/databasebuilder.h"
#include "AlbumCover/coverthread.h"

#include "Widget/statuswidget.h"

#include "debug.h"

ThreadManager* ThreadManager::INSTANCE = 0;
/*
********************************************************************************
*                                                                              *
*    Class ThreadManager                                                       *
*                                                                              *
********************************************************************************
*/
ThreadManager::ThreadManager()
{
    INSTANCE = this;

    // QThread
    _databaseBuilder     = new DataBaseBuilder();
    _collectionPopulator = new CollectionPopulator();
    _plsFilePopulator    = new PlsFilePopulator();
    _coverThread         = 0;

    // connection
    QObject::connect(_databaseBuilder,SIGNAL(buildingFinished()),this,SLOT(dbBuildFinish()));
    QObject::connect(_databaseBuilder,SIGNAL(buildingProgress(int)),this,SLOT(dbBuildProgressChanged(int)));

    QObject::connect(_collectionPopulator,SIGNAL(populatingFinished()),this,SLOT(modelCollectionPopulated()));
    QObject::connect(_collectionPopulator,SIGNAL(populatingProgress(int)),this,SLOT(modelCollectionPopulatingChanged(int)));

    QObject::connect(_plsFilePopulator,SIGNAL(populatingFinished()),this,SLOT(modelPlaylistPopulated()));
    QObject::connect(_plsFilePopulator,SIGNAL(populatingProgress(int)),this,SLOT(modelPlaylistPopulatingChanged(int)));
}


ThreadManager::~ThreadManager()
{
    Debug::debug() << "ThreadManager -> wait to finish ...";
    delete _databaseBuilder;
    delete _collectionPopulator;
    delete _plsFilePopulator;
}


/*******************************************************************************
    ThreadManager::stopThread
*******************************************************************************/
void ThreadManager::stopThread()
{
    Debug::debug() << "ThreadManager -> stop running thread";
    if(_databaseBuilder->isRunning())
      cancelThread(DB_THREAD);

    if(_collectionPopulator->isRunning())
      cancelThread(POPULATOR_C_THREAD);

    if(_plsFilePopulator->isRunning())
      cancelThread(POPULATOR_P_THREAD);

    if(_coverThread != 0)
      if(_coverThread->isRunning())
        cancelThread(COVER_SEARCH_THREAD);
}


/*******************************************************************************
    Database Scanner Thread
*******************************************************************************/
void ThreadManager::databaseBuild(QStringList listDir)
{
    if(_databaseBuilder->isRunning())
      cancelThread(DB_THREAD);

    Debug::debug() << " ThreadManager start a database builder thread";
    _databaseBuilder->rebuildFolder(listDir);
    _databaseBuilder->start();

    uint i = StatusWidget::instance()->startProgressMessage(tr("Updating Music Collection") + " (0%)");
    messageIds.insert("DbUpdate", i);
    emit dbBuildStart();    
}

void ThreadManager::dbBuildProgressChanged(int progress)
{
    QString message = QString(tr("Updating Music Collection") + " (%1%)").arg(QString::number(progress));
    StatusWidget::instance()->updateProgressMessage( messageIds.value("DbUpdate"), message );
}

void ThreadManager::dbBuildFinish()
{
    if (messageIds.contains("DbUpdate"))
      StatusWidget::instance()->stopProgressMessage( messageIds.take("DbUpdate") );

    emit dbBuildFinished();
}

bool ThreadManager::isDbRunning()
{
    return _databaseBuilder->isRunning();
}


/*******************************************************************************
    Collection Model Population thread
*******************************************************************************/
void ThreadManager::populateCollectionModel()
{
   Debug::debug() << "ThreadManager -> start a new collection populator";
   if(_collectionPopulator->isRunning()) {
     cancelThread(POPULATOR_C_THREAD);
   }

   _collectionPopulator->start();

   uint i = StatusWidget::instance()->startProgressMessage(tr("Loading Music Collection"));
   messageIds.insert("LoadMusic", i);
}

void ThreadManager::modelCollectionPopulatingChanged(int progress)
{
    QString message = QString(tr("Loading Music Collection") + " (%1%)").arg(QString::number(progress));
    StatusWidget::instance()->updateProgressMessage( messageIds.value("LoadMusic"), message );
}

void ThreadManager::modelCollectionPopulated()
{
    if (messageIds.contains("LoadMusic"))
      StatusWidget::instance()->stopProgressMessage( messageIds.take("LoadMusic") );

    emit modelPopulationFinished(MODEL_COLLECTION);

    // for each collection update do PlsFileModel update
    populatePlsFileModel();
}

/*******************************************************************************
    Playlist Model Population thread
*******************************************************************************/
void ThreadManager::populatePlsFileModel()
{
   Debug::debug() << "ThreadManager -> start a new playlist populator";
   if(_plsFilePopulator->isRunning())
     cancelThread(POPULATOR_P_THREAD);

   _plsFilePopulator->start();

   uint i = StatusWidget::instance()->startProgressMessage(tr("Loading Playlist File"));
   messageIds.insert("LoadPlaylist", i);
}

void ThreadManager::modelPlaylistPopulatingChanged(int progress)
{
    QString message = QString(tr("Loading Playlist File") + " (%1%)").arg(QString::number(progress));
    StatusWidget::instance()->updateProgressMessage( messageIds.value("LoadPlaylist"), message );
}

void ThreadManager::modelPlaylistPopulated()
{
    Debug::debug() << "ThreadManager -> modelPlaylistPopulated";

    if (messageIds.contains("LoadPlaylist"))
      StatusWidget::instance()->stopProgressMessage( messageIds.take("LoadPlaylist") );

    emit modelPopulationFinished(MODEL_PLAYLIST);
}

/*******************************************************************************
  Cover Search Thread
*******************************************************************************/
void ThreadManager::startCoverSearch()
{
    if(!_coverThread) {
      _coverThread = new CoverThread();

      connect(_coverThread,SIGNAL(finished()),this,SLOT(coverSearchFinished()));
      connect(_coverThread,SIGNAL(progress(int)),this,SLOT(coverSearchProgress(int)));
    }

    if(_coverThread->isRunning())
      cancelThread(COVER_SEARCH_THREAD);

    _coverThread->start();

    uint i = StatusWidget::instance()->startProgressMessage(tr("Fetching Album Cover"));
    messageIds.insert("CoverUpdate", i);
}

void ThreadManager::startCoverSearch(const QString& artist, const QString& album)
{
    Debug::debug() << "ThreadManager -> startCoverSearch";

    if(!_coverThread) {
      _coverThread = new CoverThread();

      connect(_coverThread,SIGNAL(finished()),this,SLOT(coverSearchFinished()));
      connect(_coverThread,SIGNAL(progress(int)),this,SLOT(coverSearchProgress(int)));
    }

    if(_coverThread->isRunning())
      cancelThread(COVER_SEARCH_THREAD);

    _coverThread->setRequest(artist, album);
    _coverThread->start();
}


void ThreadManager::coverSearchFinished()
{
    Debug::debug() << "ThreadManager -> coverSearchFinished";

    if (messageIds.contains("CoverUpdate"))
      StatusWidget::instance()->stopProgressMessage( messageIds.take("CoverUpdate") );
}

void ThreadManager::coverSearchProgress(int progress)
{
    QString message = QString(tr("Fetching Album Cover") + " (%1%)").arg(QString::number(progress));
    StatusWidget::instance()->updateProgressMessage( messageIds.value("CoverUpdate"), message );
}

/*******************************************************************************
  Cancel Thread
*******************************************************************************/
void ThreadManager::cancelThread(E_THREAD thread)
{
    switch(thread) {
      case DB_THREAD:
         Debug::warning() << "ThreadManager -> cancel database builder thread !!";
        _databaseBuilder->setExit(true);
        _databaseBuilder->wait();      // wait end of thread should be fast
        _databaseBuilder->setExit(false);
        if (messageIds.contains("DbUpdate"))
          StatusWidget::instance()->stopProgressMessage( messageIds.take("DbUpdate") );

       break;

      case POPULATOR_C_THREAD:
         Debug::warning() << "ThreadManager -> cancel collection populator thread !!";
        _collectionPopulator->setExit(true);
        _collectionPopulator->wait();      // wait end of thread should be fast
        _collectionPopulator->setExit(false);
        if (messageIds.contains("LoadMusic"))
          StatusWidget::instance()->stopProgressMessage( messageIds.take("LoadMusic") );

       break;

      case  POPULATOR_P_THREAD:
         Debug::warning() << "ThreadManager -> cancel playlist populator thread !!";
        _plsFilePopulator->setExit(true);
        _plsFilePopulator->wait();
        _plsFilePopulator->setExit(false);

        if (messageIds.contains("LoadPlaylist"))
          StatusWidget::instance()->stopProgressMessage( messageIds.take("LoadPlaylist") );

      break;

      case COVER_SEARCH_THREAD:
         Debug::warning() << "ThreadManager -> cancel cover search thread !!";
        _coverThread->setExit(true);
        if(!_coverThread->wait(2000))
          _coverThread->quit();
        _coverThread->setExit(false);
        if (messageIds.contains("CoverUpdate"))
          StatusWidget::instance()->stopProgressMessage( messageIds.take("CoverUpdate") );

      break;

      default:break;
    }
}

